from maya.cmds import *
from math import sqrt, atan2, pi
from re import findall

topology = []

# Source/target manipulation

def getObjects() :
	''' Return source and target dictionary'''
	return {'source':textField('source',q=1,tx=1),'target':textField('target',q=1,tx=1)}

def pickObject(type) :
	''' Call defineObject() when an object is selected.
	type (str) -- the type of object picked. Valid values are "source" or "target".
	'''
	select(cl=True)
	scriptJob( ct=['SomethingSelected','defineObject("'+type+'")'], protected=True, runOnce=True)
	
def defineObject(type) :
	''' If selected object is valid, define he is object type. 
	type (str) -- the type of object picked. Valid values are "source" or "target".
	'''
	objects = getObjects()
	#define the other type and the selected object
	for i in objects :
		if i != type : otherType = i
	lst = ls(sl=1)
	if lst :
		selectedObject = lst[0]
		#check if selected object is not already the other type object
		if objects[otherType] == selectedObject : displayMsg('The source and target objects can not be the same')
		else :
			#for source, check if selected object is polygonal
			check = True
			if type == 'source' :
				try : select(objects[type]+'.vtx[0]')
				except :
					displayMsg('Source object must be a polygonal object')
					check = False
			#define selected object is object type
			if check : textField(type, edit=True, tx=selectedObject)
		select(cl=True)
	
def checkObjects(objects) :
	''' If objects exists return 1, else return 0.
	objects (dict) -- objects dictionary (cf. getObjects).
	'''
	for i in objects :
		if not objExists(objects[i]) :  return 0
	return 1
	
'''	
def liveAction() :
	''Called by target translation''
	live = checkBoxGrp('live',q=1,v1=1)
	if live : makeLevel()
'''


# Distance marker

def distanceMarker() :
	''' Create and update distance marker '''
	if checkObjects(getObjects()) :
		pDist = floatFieldGrp('pDist',q=1,v=1)
		if not objExists('keepOnFloor_distanceMarker') :
			objects = getObjects()
			xyz = xform(objects['target'],q=1,ws=1,t=1)
			circle(n='keepOnFloor_distanceMarker', nr=(0,1,0), c=(xyz[0],xyz[1],xyz[2]), r=pDist)
		else :
			lst = ls(type='makeNurbCircle')
			setAttr(lst[len(lst)-1]+'.r',pDist)


# Topology

def saveTopology() :
	''' Store source topology (vector array) in global variable '''
	global topology
	objects = getObjects()
	if objExists(objects['source']) :
		clearTopology()
		i, lastVtxCoord = 0,0
		while(1) :
			vtxCoord = xform(objects['source']+'.vtx['+str(i)+']',q=1,ws=1,t=1)
			if lastVtxCoord == vtxCoord : break
			topology.append(vtxCoord)
			lastVtxCoord = vtxCoord
			i+=1
		textField('topo',edit=True,tx='preloaded')
	else : displayMsg('The source object must first be defined')
	print topology

def clearTopology() :
	''' Clear topology variable and topology textField  '''
	global topology
	topology = []
	textField('topo',edit=True,tx='')


# Make level procedures

def makeLevel(msg=False) :
	''' Check objects, load parameters and according to this 
	call placeTarget() or makeLevelOnRange().
	
	msg (bool) -- display a message if an object is missing.
	
	'''
	objects = getObjects()
	if checkObjects(objects) : 
		pNbr  = intFieldGrp('pNbr',q=1,v1=1)
		pDist = floatFieldGrp('pDist',q=1,v1=1)
		spacing = floatFieldGrp('spacing',q=1,v1=1)
		# Make level on range or just place target
		if checkBoxGrp('anim',q=1,v1=1) : makeLevelOnRange(objects, pDist, pNbr, spacing)
		else : placeTarget(objects, pDist, pNbr, spacing)
	elif msg : displayMsg('The source and target objects must first be defined')

def makeLevelOnRange(objects, pDist, pNbr, spacing) :
	''' Load animation parameters, compute what frames should be keyed and placeAndSetKeys().
	
	objects (dict) -- objects dictionary (cf. getObjects).
	pDist (float) -- the maximum vertex distance from target. Vertex more distant are not included in the calculation.
	pNbr (int) -- the maximum number of vertex included in the calculation.
	spacing (float) -- the spacing between source and target, after the leveling.
	
	'''
	def placeAndSetKeys(objects, pDist, pNbr, spacing, frame, attrs) :
		''' Call placeTarget() and set a key on each specified target attribute '''
		currentTime(frame)
		placeTarget(objects, pDist, pNbr, spacing)
		for i in attrs :
			if attrs[i] : setKeyframe(objects['target'], attribute=i, t=frame)
	
	ran = intFieldGrp('range',q=1,v=1)
	step  = intFieldGrp('step',q=1,v1=1)
	stationnary = floatFieldGrp('stationnary',q=1,v1=1)
	attrs = {checkBoxGrp('attrs',q=1,l1=1):checkBoxGrp('attrs',q=1,v1=1),
			 checkBoxGrp('attrs',q=1,l2=1):checkBoxGrp('attrs',q=1,v2=1),
			 checkBoxGrp('attrs',q=1,l3=1):checkBoxGrp('attrs',q=1,v3=1),
			 checkBoxGrp('attrs2',q=1,l1=1):checkBoxGrp('attrs2',q=1,v1=1),
			 checkBoxGrp('attrs2',q=1,l2=1):checkBoxGrp('attrs2',q=1,v2=1),
			 checkBoxGrp('attrs2',q=1,l3=1):checkBoxGrp('attrs2',q=1,v3=1)}
	
	lastCoord,keys,frames = 0,0,0
	framesList = []
	if not stationnary :
		#keyframe on each step of range
		for frame in range(ran[0], ran[1]+1, step) :
			placeAndSetKeys(objects, pDist, pNbr, spacing, frame, attrs)
			framesList.append(frame)
			frames+=1
	else :
		#keyframe only when target is stationnary
		#pass 1 (find frames)
		for frame in range(ran[0], ran[1]+1, step) :
			currentTime(frame)
			coord = geattrs(objects['target']+'.translate')
			coord = coord[0]
			if lastCoord :
				dist = distance3D(coord,lastCoord)
				if dist <= stationnary :
					framesList.append(frame)
					if not frame-1 in framesList : framesList.append(frame-1)
			lastCoord = coord
			frames+=1
		print framesList
		#pass 2 (create keyFrames)
		for frame in framesList : placeAndSetKeys(objects, pDist, pNbr, spacing, frame, attrs)
	
	displayMsg(str(frames)+' frames calculated, '+str(len(framesList))+' keys created')


# Place target algorythms

def distance2D(a,b) :
	''' Return distance in XZ 2D plane between two 3D points '''
	return sqrt( pow(a[0]-b[0], 2) + pow(a[2]-b[2], 2) )
	
def distance3D(a,b) :
	''' Return distance between two 3D points '''
	return sqrt( pow(a[0]-b[0], 2) + pow(a[1]-b[1], 2) + pow(a[2]-b[2], 2) )
	
def vertexNormal(object, vertex = ''):
	''' Return average normals of object vertex.
	
	object (str) -- the name of the object.
	vertex (int) -- the ID of a vertex.
	
	If vertex is specified function return the vertex normal (vector),
	else return all vertex normals of specified object (vector array).
	
	'''
	select(object)
	
	# Load faces normals
	facesNormal = {}
	facesInfos = polyInfo(fn=True)
	for data in facesInfos: 
		normal = findall('-?\d+\.\d{6}', data)
		id = findall('(\d+):', data)
		facesNormal[id[0]] = normal

	# Find faces connected to vertex
	if vertex != '': select(object + '.vtx[' + str(vertex) + ']')
	vertexNormal = {}
	vertexInfos = polyInfo(vertexToFace=True)
	for data in vertexInfos:
		facesList = findall('(\d) ', data)
		facesNbr = len(facesList)
		vertexId = findall('(\d+):', data)

		# Compute vertex normal
		# Vertex normal is average normal of faces connected to him
		normal = [.0, .0, .0]
		for id in facesList:
			for i in range(3):
				normal[i] += float(facesNormal[id][i])
		for i in range(3):
			normal[i] /= facesNbr
		vertexNormal[int(vertexId[0])] = normal
	
	# Clear and return vector or vector array
	select(clear=True)
	if vertex == '':
		return vertexNormal
	else:
		return vertexNormal[vertex]

def placeTarget(objects, pDist, pNbr, spacing) :
	''' Define position and rotation of target for place it on source, according to the source topology.
	
	objects (dict) -- objects dictionary (cf. getObjects).
	pDist (float) -- the maximum vertex distance from target. Vertex more distant are not included in the calculation.
	pNbr (int) -- the maximum number of vertex included in the calculation.
	spacing (float) -- the spacing between source and target, after the leveling.
	
	'''
	global topology
	targetCoord = xform(objects['target'],q=1,ws=1,t=1)

	pList = []
	if len(topology) :
		#store data from points of preloaded topology that are in pDist range
		for point in topology :
			dist = distance2D(point,targetCoord)
			if dist <= pDist :
				pList.append((dist,point[1],i))
	else :
		#store data from points of source that are in pDist range
		i, lastVtxCoord = 0,0
		while(1) :
			vtxCoord = xform(objects['source']+'.vtx['+str(i)+']',q=1,ws=1,t=1)
			if lastVtxCoord == vtxCoord : break
			dist = distance2D(vtxCoord,targetCoord)
			if dist <= pDist :
				pList.append((dist,vtxCoord[1],i))
			lastVtxCoord = vtxCoord
			i+=1
		
	#keep only the "pNbr" closest points
	pList.sort()
	if len(pList) > pNbr :
		del pList[pNbr:]
	pNbr = len(pList)
	
	if pNbr :
		# Compute vertex wheights by distance to target : ratio[x] = Ix / iSum
		# Ix = vSum - Vx
		# vSum : sum of distance values V
		# Vx : distance value of point x (pList[x][0])
		# iSum : sum of distance inverts I
		vSum = 0
		for i in range(pNbr) :
			vSum += pList[i][0]
		invert = []
		for i in range(pNbr) :
			invert.append(vSum-pList[i][0])
		if pNbr != 1 :
			iSum = 1.0 / sum(invert)
			ratio = []
			for i in range(pNbr) :
				ratio.append(invert[i]*iSum)
		else : ratio = [1]
		
		# Compute rotation
		if checkBoxGrp('rotate', q=1, v1=1):
			# Get every normal from vertex in pList
			normals = []
			for vtx in pList:
				normals.append(vertexNormal(objects['source'], vertex=vtx[2]))
			# Define target normal by weighted points normals
			normal = [.0, .0, .0]
			for i in range(pNbr):
				for j in range(3):
					normal[j] += normals[i][j]*ratio[i]
			# Calculate an Euler rotation (yaw and pitch) from the vector, and rotate the target
			yaw = atan2(normal[0], normal[1])
			pitch = atan2(normal[2], sqrt((normal[0] * normal[0]) + (normal[1] * normal[1])))
			degyaw = yaw * 180 / pi
			degpitch = pitch * 180 / pi
			if normal[0] < 0 :
				degpitch = -degpitch
			if normal[2] < 0 :
				degyaw = -degyaw
			rotate(-degpitch, 0, -degyaw, objects['target'])

		# Compute position
		# Define target Y position by weighted points heights
		y = 0
		for i in range(pNbr) :
			y += pList[i][1]*ratio[i]
		setAttr(objects['target']+'.ty',y)
		# Spacing
		if optionMenuGrp('spacingType', q=1, v=1) == 'local' :
			move(0, spacing, 0, objects['target'], relative=True, objectSpace=True)
		else :
			setAttr(objects['target']+'.ty',y+spacing)
		


# UI

def switchFooterImgs(dir) :
	''' Change main window footer icons for "make level once icon" or "make level in range icon" '''
	if checkBoxGrp('anim', q=True, v1=True) : img = 'makeLevelAnim.bmp'
	else : img = 'makeLevel.bmp'
	for i in (1,2) : picture('footerImg'+str(i), edit=True, image=dir+img)

def keepOnFloorUI(objects={'source':'','target':''}) :
	''' keepOnFloor user interface '''
	#vars
	global topology
	win = {	'name'	: 'kof',
			'title'	: 'KeepOnFloor'}
	layout = {	'w' : 400,
				'w2': 200,
				'h'	: 32,
				'm' : 3}
	scriptDir = internalVar(userAppDir=True)+'scripts/keepOnFloor/'
	
	# Reloading
	if window(win['name'], query=True, exists=True) : deleteUI(win['name'], window=True)
	if windowPref(win['name'], exists=True) : windowPref(win['name'], remove=True)
	window(win['name'], t=win['title'], maximizeButton=False, sizeable=False)
	
	#menu
	menuBarLayout()
	for i in objects :
		menu(l=i)
		menuItem(l='Pick', c='pick'+i+'()')
		menuItem(l='Clear', c='textField("'+i+'",edit=True,tx="")')
	
	menu(l='make level', pmc='makeLevel()')
	
	menu(l='?',helpMenu=True)
	menuItem(label='Help')
	menuItem(label='About KeepOnFloor', c='keepOnFloorAbout()')

	#body
	columnLayout(h=40)
	picture(image=scriptDir+'header.bmp', h=40)
	frameLayout(bv=False, lv=False, marginWidth=layout['m'])
	pLay = columnLayout(adj=True,rowSpacing=layout['m'])
	
	pickW = layout['w2']-layout['h']-2
	fieldW = layout['w2']-1.5*layout['h']
	clearW = 1.5*layout['h']
	
	#pick source/target
	for i in objects :
		rowLayout(p=pLay, nc=4, cw4=[layout['h']+2,pickW,fieldW,clearW],cl4=['center','center','center','center'],h=layout['h'])
		picture(image=scriptDir+'pick'+i+'.bmp')
		button(l='Pick '+i, c='pickObject("'+i+'")', w=pickW, h=layout['h'])
		if objects[i] : textField(i, en=False, tx=objects[i], w=fieldW, h=layout['h']+1)
		else : textField(i, en=False, tx='', w=fieldW, h=layout['h']+1)
		button(l='Clear', w=clearW, h=layout['h'], c='textField("'+i+'",edit=True,tx="")')
	#save source topology
	rowLayout(p=pLay, nc=4, cw4=[layout['h']+2,layout['w2']+2*layout['h']-2,layout['w2']-4.5*layout['h'],clearW],cl4=['center','center','center','center'])
	picture(image=scriptDir+'saveTopology.bmp',w=layout['h'])
	button(l='Preload source topology (for static source only)', c='saveTopology()', w=layout['w2']+2*layout['h']-2)
	if topology : textField('topo', en=False, tx=topo, w=layout['w2']-4.5*layout['h'])
	else : textField('topo', en=False, tx='', w=layout['w2']-4.5*layout['h'])
	button(l='Clear', w=clearW, c='clearTopology()')
	
	
	#parameters
	rowLayout(p=pLay, cw2=[layout['h'],layout['w2']])
	frameLayout(l='Parameters', borderStyle='etchedOut', w=layout['w'], mh=5)
	paramsLay = columnLayout()
	'''
	#live
	scriptJob(attributeChange=[target+'.translate',makeLevel],runOnce=True)
	scriptJob(attributeChange=[source+'.translate',makeLevel],runOnce=True)
	checkBoxGrp('live', l='live', v1=0)
	'''
	#tweaks
	rowLayout(p=paramsLay, nc=2, cw2=[layout['w2']+20,layout['w2']])
	floatFieldGrp('pDist', v1=2, l='computed area')
	intFieldGrp('pNbr', v1=2, extraLabel='max points computed', cw2=[60,150])
	rowLayout(p=paramsLay, nc=2, cw2=[layout['w2']+20,layout['w2']])
	floatFieldGrp('spacing', v1=0, l='spacing')
	optionMenuGrp('spacingType')
	menuItem(l='global')
	menuItem(l='local')
	checkBoxGrp('rotate', p=paramsLay, label='rotate', v1=1)
	
	#animation
	rowLayout(p=pLay, cw2=[layout['h'],layout['w2']])
	frameLayout(l='Animation', borderStyle='etchedOut', w=layout['w'], mh=5)
	columnLayout()
	#switch/range/step/stationnary
	checkBoxGrp('anim', label='make level on range', v1=0, changeCommand='switchFooterImgs("'+scriptDir+'")')
	checkBoxGrp('attrs', label='put keys on translation', numberOfCheckBoxes=3, labelArray3=['tx','ty','tz'], valueArray3=[0,1,0], columnWidth4=[140,80,80,80])
	checkBoxGrp('attrs2', label='put keys on rotation', numberOfCheckBoxes=3, labelArray3=['rx','ry','rz'], valueArray3=[1,1,1], columnWidth4=[140,80,80,80])
	intFieldGrp('range', numberOfFields=2, v1=playbackOptions(q=1,minTime=1), v2=playbackOptions(q=1,maxTime=1), label='start frame', extraLabel='stop frame', cw=[4,100])
	intFieldGrp('step', v1=playbackOptions(q=1,by=1), label='step')
	floatFieldGrp('stationnary', v1=0, l='only motionless frames', extraLabel='motionless tolerance (0 = off)', cw=[3,150])
	#end frame
	setParent('..')
		
	#make level
	rowLayout(p=pLay, nc=3, cw3=[layout['h']+2,layout['w']-2*layout['h']-3,layout['h']],cl3=['center','center','center'],h=layout['h'])
	picture('footerImg1',image=scriptDir+'makeLevel.bmp')
	button(l='Make level', c='makeLevel(msg=1)', w=layout['w']-2*layout['h']-5, h=layout['h'])
	picture('footerImg2',image=scriptDir+'makeLevel.bmp')
		
	window(win['name'], edit=True, resizeToFitChildren=True)
	showWindow(win['name'])

def displayMsg(txt) :
	''' Simple window for message display '''
	confirmDialog(title='Message', button = 'Ok', message=txt)
	
def keepOnFloorAbout() :
	''' Decorated window for "About" display '''
	name = 'kof_about'
	#creation/reset
	if window(name, exists=True) : deleteUI(name, window=True)
	if windowPref(name, exists=True) : windowPref(name, remove=True)
	window(name, title='About...', maximizeButton=False, sizeable=False, resizeToFitChildren=True)
	#body
	columnLayout(adj=1)
	picture(image=internalVar(userAppDir=True)+'scripts/keepOnFloor/about.bmp', width=146, height=154)
	showWindow(name)


# Run

keepOnFloorUI()